# time-otago.R
#
# Example from Ross Ihaka's talk:
# "The R Project: A Brief History and Thoughts About the Future"
# http://www.stat.auckland.ac.nz/~ihaka/downloads/Otago.pdf

library(jit)
XLEN <- 3000
YLEN <- XLEN
set.seed(1994)
x = seq(0, 1, by = 1/(XLEN-1))
y = runif(YLEN)

otago.cute <- function()
{
    x[apply(outer(x, y, function(x, y) abs(y - x)),
            2, function(u) which(u == min(u))[1])]
}
otago.obvious <- function()
{
    nx = numeric(length(y))
    for(j in 1:length(y)) {
        dmin = Inf
        imin = 0L # 0L not 0 for consistency with otago.jit1
        for(i in 1:length(x)) {
            d = abs(x[i] - y[j])
            if (d < dmin) {
                dmin = d
                imin = i
            }
        }
        nx[j] = x[imin]
    }
    nx
}
otago.jit1 <- function()
{
    jit(1)
    nx = numeric(length(y))
    for(j in 1:length(y)) {
        dmin = Inf
        imin = 0L # 0L needed to prevent type change below
        for(i in 1:length(x)) {
            d = abs(x[i] - y[j])
            if (d < dmin) {
                dmin = d
                imin = i
            }
        }
        nx[j] = x[imin]
    }
    nx
}
otago.jit2 <- function()
{
    jit(2) # experimental
    nx = numeric(length(y))
    for(j in 1:length(y)) {
        dmin = Inf
        imin = 0L
        for(i in 1:length(x)) {
            d = abs(x[i] - y[j])
            if (d < dmin) {
                dmin = d
                imin = i
            }
        }
        nx[j] = x[imin]
    }
    nx
}
do.time <- function(func)
{
    time <- system.time(results <- func())[3]
    func.name <- deparse(substitute(func))
    if(func.name == "otago.cute") {
        reference.time <<- time # NOTE: <<- not <-
        reference.results <<- results
    }
    stopifnot(!is.na(reference.time))
    stopifnot(identical(reference.results, results))
    cat(sprintf("%-14.14s %7.3f    %7.2f      %7.2f\n",
        func.name,
        time,
        time / reference.time,
        reference.time / time))
    time
}
cat(version$version.string, "\n\n")
cat("XLEN=", XLEN, " YLEN=", YLEN, "\n\n", sep="")
cat("Big \"relative\" means the given function is slow relative to otago.cute\n\n")
cat("function          time   relative   1/relative\n")
reference.time <- NA
reference.results <- NA
do.time(otago.cute)
do.time(otago.obvious)
do.time(otago.jit1)
do.time(otago.jit2)
